import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from IPython.display import HTML, display
import ipywidgets as widgets

class RootFinderOpen:
    def __init__(self, f, fprime=None, x0=None, x1=None, tol=1e-6, max_iter=20):
        self.f = f
        self.fprime = fprime
        self.x0 = x0
        self.x1 = x1
        self.tol = tol
        self.max_iter = max_iter
        self.newton_guesses = self._run_newton()
        self.secant_guesses = self._run_secant()

    def _run_newton(self):
        if self.fprime is None or self.x0 is None:
            return []
        x = self.x0
        guesses = [x]
        for _ in range(self.max_iter):
            fx, fpx = self.f(x), self.fprime(x)
            if abs(fpx) < 1e-10:
                break
            x_new = x - fx / fpx
            guesses.append(x_new)
            if abs(x_new - x) < self.tol:
                break
            x = x_new
        return guesses

    def _run_secant(self):
        if self.x0 is None or self.x1 is None:
            return []
        x0, x1 = self.x0, self.x1
        guesses = [x0, x1]
        for _ in range(self.max_iter):
            f0, f1 = self.f(x0), self.f(x1)
            if abs(f1 - f0) < 1e-10:
                break
            x2 = x1 - f1 * (x1 - x0) / (f1 - f0)
            guesses.append(x2)
            if abs(x2 - x1) < self.tol:
                break
            x0, x1 = x1, x2
        return guesses

    def _make_animation_open(self, method='Newton', interval_ms=750):
        if method == 'Newton':
            guesses = self.newton_guesses
            title_method = "Newton's Method"
            line_label = "Tangent Line"
        else:
            guesses = self.secant_guesses
            title_method = "Secant Method"
            line_label = "Secant Line"

        if not guesses or len(guesses) < 2:
            return None

        total_steps = len(guesses)
        min_g, max_g = min(guesses), max(guesses)

        x_span = max_g - min_g if max_g > min_g else 1.0
        x_pad = max(0.2 * x_span, 1.0)
        x_min, x_max = min_g - x_pad, max_g + x_pad

        x_vals = np.linspace(x_min, x_max, 400)
        y_vals = self.f(x_vals)

        fig, ax = plt.subplots(figsize=(10, 6))
        ax.plot(x_vals, y_vals, label='f(x)', color='blue')
        ax.axhline(0, color='gray', linestyle='--', alpha=0.7)

        line, = ax.plot([], [], marker='o', linestyle='-', color='red', markersize=8, label='Iterations')
        aux_line, = ax.plot([], [], color='green', linestyle='--', label=line_label)
        current_point, = ax.plot([], [], marker='o', color='gold', markersize=10, markeredgecolor='red', label='Current $x_k$')
        vertical_line, = ax.plot([], [], color='purple', linestyle=':', linewidth=2, label='Next guess connector')

        ax.set_title(f"{title_method} Animation", fontsize=14)
        ax.set_xlabel("x", fontsize=12)
        ax.set_ylabel("f(x)", fontsize=12)
        ax.grid(True, linestyle=':', alpha=0.6)
        ax.legend()
        ax.set_xlim(x_min, x_max)

        f_g = [self.f(g) for g in guesses]
        y_min, y_max = min(f_g), max(f_g)

        y_span = y_max - y_min if y_max > y_min else 1.0
        y_pad = max(0.2 * y_span, 1.0)
        ax.set_ylim(y_min - y_pad, y_max + y_pad)

        def update_open(frame):
            step = frame
            shown_guesses = guesses[:step+1]
            f_guesses = [self.f(g) for g in shown_guesses]

            line.set_data(shown_guesses, f_guesses)
            current_point.set_data([shown_guesses[-1]], [f_guesses[-1]])
            aux_line.set_data([], [])
            vertical_line.set_data([], [])
            ax.set_title(f"{title_method} - Step {step}/{total_steps-1}", fontsize=14)

            if step > 0:
                x_k_prev = shown_guesses[step - 1]
                f_k_prev = self.f(x_k_prev)

                if method == 'Newton':
                    m = self.fprime(x_k_prev)
                elif method == 'Secant' and step > 1:
                    x_k_minus_2 = shown_guesses[step - 2]
                    f_k_minus_2 = self.f(x_k_minus_2)
                    m = (f_k_prev - f_k_minus_2) / (x_k_prev - x_k_minus_2) if abs(x_k_prev - x_k_minus_2) > 1e-10 else None
                else:
                    m = None

                if m is not None:
                    x_line = np.linspace(x_min, x_max, 400)
                    y_line = f_k_prev + m * (x_line - x_k_prev)
                    aux_line.set_data(x_line, y_line)
                    if abs(m) > 1e-12:
                        x_intersect = x_k_prev - f_k_prev / m
                        vertical_line.set_data([x_intersect, x_intersect], [0, self.f(shown_guesses[step])])

            return line, current_point, aux_line, vertical_line

        anim = FuncAnimation(
            fig, update_open,
            frames=total_steps,
            interval=interval_ms,
            blit=True,
            repeat=False
        )
        plt.close(fig)
        return HTML(anim.to_jshtml())

    def show_toggle_open(self, interval_ms=750):
        newton_anim_html = self._make_animation_open('Newton', interval_ms)
        secant_anim_html = self._make_animation_open('Secant', interval_ms)

        if newton_anim_html is None or secant_anim_html is None:
            print("Cannot create animations. Check function definition or initial points (x0, x1).")
            return

        newton_content = widgets.HTML(value=newton_anim_html.data)
        secant_content = widgets.HTML(value=secant_anim_html.data)

        tab_container = widgets.Tab()
        tab_container.children = [newton_content, secant_content]

        tab_container.set_title(0, 'Newton')
        tab_container.set_title(1, 'Secant')

        display(tab_container)



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from IPython.display import HTML, display
import ipywidgets as widgets

class RootFinderClosed:
    def __init__(self, f, a=None, b=None, tol=1e-6, max_iter=20):
        self.f = f
        self.a = a
        self.b = b
        self.tol = tol
        self.max_iter = max_iter
        self.bisect_intervals, self.bisect_guesses = self._run_bisection()
        self.falsepos_intervals, self.falsepos_guesses = self._run_false_position()

    def _run_bisection(self):
        if self.a is None or self.b is None:
            return [], []
        a, b = self.a, self.b
        if self.f(a) * self.f(b) > 0:
            return [], []

        intervals = []
        guesses = []
        for _ in range(self.max_iter):
            c = (a + b) / 2.0
            intervals.append((a, b, c))
            guesses.append(c)
            if abs(self.f(c)) < self.tol or abs(b - a) < self.tol:
                break
            if self.f(a) * self.f(c) < 0:
                b = c
            else:
                a = c
        return intervals, guesses

    def _run_false_position(self):
        if self.a is None or self.b is None:
            return [], []
        a, b = self.a, self.b
        if self.f(a) * self.f(b) > 0:
            return [], []

        intervals = []
        guesses = []
        for _ in range(self.max_iter):
            fa, fb = self.f(a), self.f(b)
            if abs(fb - fa) < 1e-12:
                break
            c = b - fb * (b - a) / (fb - fa)
            intervals.append((a, b, c))
            guesses.append(c)
            if abs(self.f(c)) < self.tol:
                break
            if fa * self.f(c) < 0:
                b = c
                fb = self.f(b)
            else:
                a = c
                fa = self.f(a)
        return intervals, guesses

    def _make_animation_closed(self, method='Bisection', interval_ms=750):
        if method == 'Bisection':
            intervals, guesses = self.bisect_intervals, self.bisect_guesses
            title_method = "Bisection Method"
            guess_label = "Midpoint"
        else:
            intervals, guesses = self.falsepos_intervals, self.falsepos_guesses
            title_method = "False Position Method"
            guess_label = "Point c"

        if not guesses:
            return None

        total_steps = len(guesses)
        all_points = [g for g in guesses] + [p for ab in intervals for p in ab[:2]]
        min_g, max_g = min(all_points), max(all_points)

        x_span = max_g - min_g if max_g > min_g else 1.0
        x_pad = max(0.2 * x_span, 1.0)
        x_min, x_max = min_g - x_pad, max_g + x_pad

        x_vals = np.linspace(x_min, x_max, 400)
        y_vals = self.f(x_vals)

        fig, ax = plt.subplots(figsize=(10, 6))
        ax.plot(x_vals, y_vals, label='f(x)', color='blue')
        ax.axhline(0, color='gray', linestyle='--', alpha=0.7)

        line, = ax.plot([], [], marker='o', linestyle='-', color='red', markersize=6, alpha=0.6, label='Previous guesses')
        a_line = ax.axvline(0, color='red', linestyle=':', lw=1.5, alpha=0.9, label='a (left endpoint)')
        b_line = ax.axvline(0, color='green', linestyle=':', lw=1.5, alpha=0.9, label='b (right endpoint)')
        current_point, = ax.plot([], [], marker='o', color='gold', markersize=10, markeredgecolor='black', label=guess_label)
        vertical_line, = ax.plot([], [], color='purple', linestyle=':', linewidth=2, label='Connector')

        ax.set_title(f"{title_method} Animation", fontsize=14)
        ax.set_xlabel("x", fontsize=12)
        ax.set_ylabel("f(x)", fontsize=12)
        ax.grid(True, linestyle=':', alpha=0.6)
        ax.legend()
        ax.set_xlim(x_min, x_max)

        f_g = [self.f(g) for g in guesses]
        y_min, y_max = min(f_g + [0]), max(f_g + [0])
        y_span = y_max - y_min if y_max > y_min else 1.0
        y_pad = max(0.2 * y_span, 1.0)
        ax.set_ylim(y_min - y_pad, y_max + y_pad)

        def update_closed(frame):
            step = frame
            shown_intervals = intervals[:step+1]
            shown_guesses = guesses[:step+1]
            f_guesses = [self.f(g) for g in shown_guesses]

            a, b, c = shown_intervals[-1]
            a_line.set_xdata([a])
            b_line.set_xdata([b])

            line.set_data(shown_guesses[:-1], f_guesses[:-1])
            current_point.set_data([c], [self.f(c)]) 
            vertical_line.set_data([c, c], [0, self.f(c)])

            ax.set_title(f"{title_method} - Step {step}/{total_steps-1}", fontsize=14)
            return line, current_point, vertical_line, a_line, b_line

        anim = FuncAnimation(
            fig, update_closed,
            frames=total_steps,
            interval=interval_ms,
            blit=True,
            repeat=False
        )
        plt.close(fig)
        return HTML(anim.to_jshtml())

    def show_toggle_closed(self, interval_ms=750):
        bisect_anim_html = self._make_animation_closed('Bisection', interval_ms)
        falsepos_anim_html = self._make_animation_closed('False Position', interval_ms)

        if bisect_anim_html is None or falsepos_anim_html is None:
            print("Cannot create animations. Check initial interval [a, b].")
            return

        bisect_content = widgets.HTML(value=bisect_anim_html.data)
        falsepos_content = widgets.HTML(value=falsepos_anim_html.data)

        tab_container = widgets.Tab()
        tab_container.children = [bisect_content, falsepos_content]

        tab_container.set_title(0, 'Bisection')
        tab_container.set_title(1, 'False Position')

        display(tab_container)



from setuptools import setup

setup(
    name='Mint_NM',
    version='0.1.26',
    description='Tools for ENGPHYS 3NM4',
    url='https://github.com/mwelland/themintlab/Mint_NM',
    author='Michael Welland',
    packages=['Mint_NM'],
    install_requires=[
        'numpy',
        'matplotlib',
        'ipywidgets',
        'IPython',
        'pyppeteer',
        'nbconvert',
        'scipy',
        'plotly'
    ],
    python_requires='>=3.7',
    license="CC-BY-4.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
    ],
)
